/*
 *  TOPPERS/FMP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Flexible MultiProcessor Kernel
 *
 *  Copyright (C) 2015 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *
 *  上記著作権者は，以下の(1)～(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 *
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 *
 */
#ifndef TOPPERS_IMX6_H
#define TOPPERS_IMX6_H

#include <sil.h>

/*
 * CPU clocks
 */
#define IMX6_CORE_CLK           800U
#define IMX6_PERIPHCLK          (IMX6_CORE_CLK/2U)
#define MPCORE_GTM_CLOCK_FREQ_MHZ  IMX6_PERIPHCLK

/*
 * Tmer preescaler and load value to achieve 1ms per tick.
 * Note that the preescaler value must be representable with 8 bits;
 * and the load value must be a 32bit value.
 *
 * Private timers and watchdogs are clocked with PERIPHCLK which is 1/2 of
 * the CPU frequency. The formula is:
 *
 *    period = (preescaler + 1)(load + 1) / PERIPHCLK
 *      - period    = 1ms
 *      - periphclk = 200MHz
 *
 * For example:
 *      - preescaler = 0
 *      - load       = 200000
 */
#define MPCORE_WDT_PS_1MS    0U                          /* preescaler for private watchdog */
#define MPCORE_WDT_LOAD_1MS  (IMX6_PERIPHCLK * 1000U) /* load value for private watchdog */

#define MPCORE_TM_PS_1MS     0U                          /* preescaler for private timer */
#define MPCORE_TM_LOAD_1MS   (IMX6_PERIPHCLK * 1000U) /* load value for private timer */

#define MPCORE_GTM_PS_1MS    0U                          /* preescaler for global timer */
#define MPCORE_GTM_LOAD_1MS  (IMX6_PERIPHCLK * 1000U) /* load value for global timer */

/*
 *  MPCore Private Memory Region Base Address
 */
#define MPCORE_PMR_BASE  0x00A00000

/*
 *  GIC Base Address
 */
#define GICC_BASE		(MPCORE_PMR_BASE + 0x0100)
#define GICD_BASE		(MPCORE_PMR_BASE + 0x1000)

/*
 *  Number of interrupts supported by the GICv1.0 in this board. Note
 *  that these values could be obtained dynamically from the corresponding
 *  GIC register.
 */
#define GIC_TNUM_INTNO		160U

/*
 *  L2(PL310) Base Address
 */
#define PL310_L2CACHE_BASE	0x00A02000

/*
 *  UART
 */
#define UART1_BASE   0x02020000
#define UART2_BASE   0x021E8000
#define UART3_BASE   0x021EC000

/*
 *  Interrupt Number
 */
#define DIC_IRQNO_UART1 58
#define DIC_IRQNO_UART2 59
#define DIC_IRQNO_UART3 60

/*
 *  System Reset Controller
 */
#define SRC_SCR  0x020D8000
#define SRC_GPR1 0x020D8020  /* CORE0 Entry */
#define SRC_GPR2 0x020D8024  /* CORE0 Arg   */
#define SRC_GPR3 0x020D8028  /* CORE1 Entry */
#define SRC_GPR4 0x020D802C  /* CORE1 Arg   */
#define SRC_GPR5 0x020D8030  /* CORE2 Entry */
#define SRC_GPR6 0x020D8034  /* CORE2 Arg   */
#define SRC_GPR7 0x020D8038  /* CORE3 Entry */
#define SRC_GPR8 0x020D803C  /* CORE3 Arg   */

#define SRC_SCR_CORE1_ENA 0x00400000
#define SRC_SCR_CORE2_ENA 0x00800000
#define SRC_SCR_CORE3_ENA 0x01000000

/*
 *  IOMUXC
 */
#define IOMUXC_BASE 0x020E0000

#define IOMUXC_PAD_CTL_PKE		(1 << 12)
#define IOMUXC_PAD_CTL_PUE		(1 << 13 | IOMUXC_PAD_CTL_PKE)
#define IOMUXC_PAD_CTL_PUS_100K_UP	(2 << 14 | IOMUXC_PAD_CTL_PUE)
#define IOMUXC_PAD_CTL_SPEED_MED	(2 << 6)
#define IOMUXC_PAD_CTL_DSE_40ohm	(6 << 3)
#define IOMUXC_PAD_CTL_SRE_FAST	(1 << 0)
#define IOMUXC_PAD_CTL_HYS		(1 << 16)

#ifndef TOPPERS_MACRO_ONLY

/*
 *  Trace log definitions (TODO: not tested)
 *
 *  TRACE_GET_TIM returns microseconds and uses the lower part of the
 *  global timer counter (32bit). This means that only around 12.89 seconds
 *  (0xFFFFFFFF/333 us) of trace can be taken until the counter wraps around.
 *
 *  Note: For traces that take more time, comment the TRACE_HW_INIT and
 *  TRACE_GET_TIM. As defined in trace_config.h, get_my_current_time()
 *  will be used instead. This allows taking a very long trace log (up to
 *  49,7 days) but the resolution will be in the milliseconds order.
 */
#ifndef MPCORE_GTC_COUNT_L
#define MPCORE_GTC_COUNT_L (MPCORE_PMR_BASE + 0x0200)
#endif  /* MPCORE_GTC_COUNT_L */

#define TRACE_HW_INIT() x_init_pcc()
#define TRACE_GET_TIM() (*((volatile uint32_t *)(MPCORE_GTC_COUNT_L)) / (MPCORE_GTC_CLOCK_FREQ_MHZ))

#endif /* TOPPERS_MACRO_ONLY */

#define IOMUX_GPR3 0x020E000C
#define IOMUX_GPR3TZASC1_LOCK (1<<11)
#define IOMUX_GPR3TZASC2_LOCK (1<<12)

#define IOMUX_GPR9 0x020E0024
#define IOMUX_GPR9_TZASC1_BYP 0x01
#define IOMUX_GPR9_TZASC2_BYP 0x02

#define TZASC1_BASE 0x021D0000
#define TZASC2_BASE 0x021D4000


#ifdef TOPPERS_CHIP_SOLOX

/*
 * i.MX6Solox(Saber)
 */

/*
 * Memory base address and size
 */
#define DDR_ADDR 0x80000000
#define DDR_SIZE 0x3fffffff /* 1000MB */
#define OSRAM_ADDR 0x00900000

#define UART_UBIR_VAL   0x000f
#define UART_UBMR_VAL   0x0068

/*
 * Default number of processors (this value can be overwritten in the
 * application's Makefile with the desired number of processors)
 */
#define TNUM_PRCID  1

#define UART1_TX_IOMUX 0x00024,0,0x0830,0,0x036C,UART_PAD_CTL_VAL
#define UART1_RX_IOMUX 0x0028,0,0x0830,1,0x0370,UART_PAD_CTL_VAL
#define UART2_TX_IOMUX 0x002C,0,0x0838,0,0x0374,UART_PAD_CTL_VAL
#define UART2_RX_IOMUX 0x0030,0,0x0838,1,0x0378,UART_PAD_CTL_VAL

#else /* !TOPPERS_CHIP_SOLOX */

/*
 * i.MX6Quad(Saber)
 */

/*
 * Memory base address and size
 */
#define DDR_ADDR 0x10000000
#define DDR_SIZE 0x3fffffff /* 1000MB */
#define OSRAM_ADDR 0x00900000

#define UART_UBIR_VAL   0x000f
#define UART_UBMR_VAL   0x015b

/*
 * Default number of processors (this value can be overwritten in the
 * application's Makefile with the desired number of processors)
 */
#ifndef TNUM_PRCID
#define TNUM_PRCID  4
#endif /* TNUM_PRCID */

#define UART1_TX_IOMUX 0x0280,3,0x0000,0,0x0650,UART_PAD_CTL_VAL
#define UART1_RX_IOMUX 0x0284,3,0x0920,1,0x0654,UART_PAD_CTL_VAL
#define UART2_TX_IOMUX 0x00BC,4,0x0000,0,0x03D0,UART_PAD_CTL_VAL
#define UART2_RX_IOMUX 0x00C0,4,0x0928,1,0x03D4,UART_PAD_CTL_VAL
#define UART3_TX_IOMUX 0x00B4,2,0x0000,0,0x03C8,UART_PAD_CTL_VAL
#define UART3_RX_IOMUX 0x00B8,2,0x0930,1,0x03CC,UART_PAD_CTL_VAL

#endif /* TOPPERS_CHIP_SOLOX */

#endif /* TOPPERS_IMX6_H */
